;;
;; Copyright (c) 1996-1998 John Zuckerman
;;
;; See the file "Notice" for information on usage and redistribution
;; of this file, and for a DISCLAIMER OF ALL WARRANTIES.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;
(require "../common/app.ss")
(require "../common/scrollframe.ss")
(require "../common/select-file.ss")
(require "../common/warning-dialog.ss")
(require "../common/auxio.ss")
(require "edit-text.ss")

(define-class (<edit-toplevel> filename) (<app-toplevel>)
  (ivars
   (current-filename filename)
   )
  (inherited mini-buffer mini-search)
  (inheritable mini-buffer mini-search)
  (private
    [set-current-filename! (name)
     (set! current-filename name)
     (set-title! self (if name (format "Editing ~s" name) "Editor"))])
  (protected)
  (public
   [init (filename)
    (send-base self init)
    (set-current-filename! filename)
    (let () ;;; needed for defns
      (define-syntax simple-menu-item
        (syntax-rules ()
          ((_ (str1 str2) act (argname argval) ...)
           (create <command-menu-item> with
                   (title: (cons str1 str2))
                   (action: act)
                   (argname argval) ...
                   ))))
      (define make-backup-file
        (lambda (name)
          (define hackslash
            (case-lambda 
              [(s) (hackslash s #\/ #\\)]
              [(s from to)
               (if (not (memv (machine-type) '(ppcnt i3nt)))
                   s
                   (let ([s (string-copy s)])
                     (let f ([i (fx- (string-length s) 1)])
                       (unless (fx< i 0)
                         (when (char=? (string-ref s i) from)
                           (string-set! s i to))
                         (f (fx- i 1))))
                     s))]))
          (define get-filename
            (lambda (str)
              (list->string
               (reverse
                (let f ((l (reverse (string->list str))))
                  (cond [(null? l) '()]
                        [(char=? (car l) #\\) '()]
                        [(char=? (car l) #\:) '()]
                        [else (cons (car l) (f (cdr l)))]))))))
          (when (file-exists? name)
            (if (memv (machine-type) '(ppcnt i3nt))
                (let ((cp
                       (string-append "copy "
                                      (hackslash name)
                                      " "
                                      (hackslash name)
                                      ".backup")))
                  (system cp)
                  (system (string-append "del " (hackslash name)))
                  )
                (system (string-append "mv -f " name " " name ".backup"))))
          ))
      ;;
      ;; widget bindings
      ;;
      (let* ((scrolled-frame
              (create <scrollframe> self with
                      (default-vscroll: #t)
                      (sticky-hscroll: #t)
                      ))
             (edit-wgt
              (create <edit-text> scrolled-frame with
                      (background-color: 'white)
                      (foreground-color: 'black)
                      (wrap: 'none)
                      ))
             (token (swl:begin-application (lambda () (send self destroy))))
             ) ;;; widget bindings

        (send edit-wgt set-font! (create <font> 'courier 12 '()))

        (send self notify-text edit-wgt)

        (send self set-destroy-request-handler!
              (lambda (self)
                (and (or (not (send edit-wgt buffer-modified?))
                         (begin
                           ; make sure the <toplevel> window is visible
                           ; and on top so we know which editor we're
                           ; about to blow away.
                           (send self show)
                           (send self raise)
                           (swl:sync-display)
                           (let ([ans
                                  (warning-dialog self
(format "
     The ~a is not saved to disk.
     Do you really want to quit?
"
(if current-filename (format "file ~s" current-filename) "edit buffer"))
                                                  '(yes no cancel))])
                             (eq? ans 'yes))))
                     (begin (swl:end-application token) #t))))
        ;;
        ;; action & menu-item bindings
        ;;
        (letrec
            ((buffer-modified-menu-item (create <command-menu-item>))
             
             (new-action (lambda (item) (new-edit)))

             (paste-action
              (lambda (item) (send edit-wgt paste)))

             (load-the-file
              (lambda (name item)
                (send item set-enabled! #f)
                (set-current-filename! name)
                (send save-menu-item set-enabled! #t)
                (let ([end (send edit-wgt get-end-mk)])
                  (swl:block-read-file
                   name
                   (lambda (string)
                     (send edit-wgt raw-insert-at end string))))
                (send edit-wgt set-cursor-pos! '(0 . 0))))
             
             (open-action
              (lambda (item)
                (send item set-enabled! #f)
                (thread-fork
                 (lambda ()
                   (let ((name (select-file
                                (current-directory) "Open" self 'existing)))
                     (if name
                         (load-the-file name item)
                         (send item set-enabled! #t)))))))

             (do-save
              (lambda ()
                (make-backup-file current-filename)
                (swl:write-file-from-widget current-filename edit-wgt)
                (set-title! buffer-modified-menu-item "")
                (send edit-wgt set-buffer-to-safe!)))

             (do-save-as
               (lambda ()
                 (let ((name (select-file (current-directory)
                                          "Save _As" self 'non-existing)))
                   (if name
                       (begin
                         (set-current-filename! name)
                         (send save-menu-item set-enabled! #t)
                         (make-backup-file current-filename)
                         (delete-file current-filename)
                         (swl:write-file-from-widget current-filename edit-wgt)
                         (set-title! buffer-modified-menu-item "")
                         (send edit-wgt set-buffer-to-safe!)
                         #t)
                       #f))))

             (save-action
              (lambda (item)
                (if current-filename
                    (do-save)
                    (thread-fork do-save-as))))

             (save-as-action
              (lambda (item)
                (thread-fork do-save-as)))

             (execute-action
              (lambda (item)
                (thread-fork
                  (lambda ()
                    (if current-filename
                        (do-save)
                        (do-save-as))
                    (when current-filename
                      (load current-filename))))))

             (new-menu-item
              (simple-menu-item ("_New" "Alt+n")
                                (lambda (item) (new-action item))))

             (open-menu-item
              (simple-menu-item ("_Open" "Alt+o")
                                (lambda (item) (open-action item))))

             (save-menu-item
              (simple-menu-item ("_Save" "Alt+s")
                                (lambda (item) (save-action item))
                                (enabled: #f)))

             (save-as-menu-item
              (simple-menu-item ("Save _as" "Alt+a")
                                (lambda (item) (save-as-action item))))

             (quit-menu-item
              (simple-menu-item ("_Quit" "Alt+q")
                                (lambda (item) (send self destroy))))

             (execute-menu-item
              (simple-menu-item ("Save and _Execute" "Alt+g")
                                (lambda (item) (execute-action item))))

             (copy-menu-item
              (simple-menu-item ("_Copy" "Alt+c")
               (lambda (item) (send edit-wgt action-copy))))

             (cut-menu-item
              (simple-menu-item ("Cu_t" "Alt+x")
               (lambda (item) (send edit-wgt action-cut))))

             (paste-menu-item
              (simple-menu-item ("_Paste" "Alt+v")
               (lambda (item) (send edit-wgt action-paste))))

             (search-forward-menu-item
              (simple-menu-item ("Search _Forward" "Ctrl+s")
                                (lambda (item) (send edit-wgt search-forward))))

             (search-backward-menu-item
              (simple-menu-item ("Search _Backward" "Ctrl+r")
                                (lambda (item) (send edit-wgt search-backward))))
             (tempcond-menu-item
              (simple-menu-item ("_cond" "Ctrl+q c")
                                (lambda (item)
                                  (send edit-wgt insert-template 'cond))))
             (tempdefine-menu-item
              (simple-menu-item ("_define" "Ctrl+q d")
                                (lambda (item)
                                  (send edit-wgt insert-template 'define))))
             (templambda-menu-item
              (simple-menu-item ("_lambda" "Ctrl+q a")
                                (lambda (item)
                                  (send edit-wgt insert-template 'lambda))))
             (templet-menu-item
              (simple-menu-item ("l_et" "Ctrl+q l")
                                (lambda (item)
                                  (send edit-wgt insert-template 'let))))
             ) ;;; action & menu-item bindings
          ;;
          ;; menu bindings
          ;;
          (let*
              ((menu-file
                 (create <cascade-menu-item> with
                   (title: "_File")
                   (menu:
                     (create <menu>
                       (list new-menu-item
                             open-menu-item
                             execute-menu-item
                             save-menu-item
                             save-as-menu-item
                             quit-menu-item)))))
               (menu-edit
                 (create <cascade-menu-item> with
                   (title: "_Edit")
                   (menu:
                     (create <menu>
                       (list copy-menu-item
                             cut-menu-item
                             paste-menu-item
                             search-forward-menu-item
                             search-backward-menu-item)))))
               #;(menu-templates
                 (create <cascade-menu-item> with
                   (title: "_Templates")
                   (menu:
                     (create <menu>
                       (list tempcond-menu-item
                             tempdefine-menu-item
                             templambda-menu-item
                             templet-menu-item)))))
               (menu-preferences 
                 (create <cascade-menu-item> with
                   (title: "P_references")
                   (menu:
                     (make-menu
                       ("_Font..."
                         (lambda (item)
                           (swl:font-dialog self
                             (swl:font-families 'fixed)
                             '(-8 -10 -12 -14 -16 -18 -20 -22 -24
                                8 10 12 14 16 18 20 22 24)
                             '(bold normal)
                             (lambda () (send edit-wgt get-font))
                             (lambda (fnt)
                               (when fnt (send edit-wgt set-font! fnt))))))
                       ("_Toggles"
                         (create <menu>
                           (list (create <check-menu-item> with
                                   (title: "The _Box")
                                   (action:
                                     (lambda (item)
                                       (send edit-wgt show-the-box)))))))))))
; Due to bug in Tk 8.0.4 on Unix, command-menu-items are never invoked
; if drawn in a menubar (ie. as menu: option of a <toplevel>).
; I have a patch to tk8.0.4/library/menu.tcl that fixes this, but instead
; plan to make Help a cascade.
;                 (create <command-menu-item> with
;                   (title: "Help")
;                   (action:
;                     (lambda (item)
;                       (let ([root (getenv "SWL_ROOT")])
;                         (if (not root)
;                             (warning-dialog self "Can't locate documentation.  (SWL_ROOT environment variable not set)" 'oops)
;                             (new-viewer (string-append root "/edit.html")))))))
               ) ;;; menu bindings
              
            (send edit-wgt set-menu-items!
                  buffer-modified-menu-item
                  new-menu-item open-menu-item
                  execute-menu-item
                  save-menu-item save-as-menu-item
                  quit-menu-item
                  copy-menu-item
                  cut-menu-item
                  paste-menu-item
                  #f
                  )

            (when current-filename
              (send open-menu-item set-enabled! #f)
              (send save-menu-item set-enabled! #t))
            (send self set-menu!
              (create <menu>
                (list menu-file
                      menu-edit
                      #;menu-templates
                      menu-preferences
                      buffer-modified-menu-item
                      (swl:help-menu)
                      (swl:application-menu)
                      )))
            ; hack (until we fix menu.ss, menu configuration has to be done
            ;       after the menu is installed via set-menu!)
            (set-relief! (send self get-menu) 'flat)
            (pack scrolled-frame (expand: #t) (fill: 'both))
            (thread-fork-group
              (lambda ()
                (when (and current-filename
                           (string? current-filename)
                           (file-exists? current-filename))
                  (load-the-file current-filename open-menu-item))))
            ;; This makes the edit-wgt responsive to key bindings on startup
            ;; we should fix SWL so this isn't needed.
            (send edit-wgt set-focus)
            )))
      (void))]
   )
  )

(define new-edit
  (case-lambda
    [() (new-edit #f)]
    [(current-filename)
     (swl:make-application
      (lambda () (create <edit-toplevel> current-filename)))
     (void)]))
 
; (swl:register-application "New Editor" new-edit)

