;; Copyright (c) 1996 Oscar Waddell
;;
;; See the file "Notice" for information on usage and redistribution
;; of this file, and for a DISCLAIMER OF ALL WARRANTIES.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Base classes for unnamed Scheme Graphics Library
;; $Id: base1.ss,v 1.11 1998/02/23 23:27:38 owaddell Exp $

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Tk-Object
;;
;;   Let's us ask if something isa? tk-object, and sanitizes widget destroy.

;; Note:
;;   destroy method currently blitzes only the public dispatch table
;;   presumably it is an error if user code calls destroy and then continues
;;   to call protected or private methods on the object.  ie. destroy is
;;   the last thing that should be done.  since this bangs on internal
;;   representation, I was tempted to put this in ../oop/class.ss, instead
;;   we might generalize this

(define-swl-class (<tk-object>) (<base>)
  ;* Documentation for this class will not be printed
  ;* using the current document compiler.
  (ivars)
  (inherited)
  (inheritable)
  (private)
  (protected)
  (public
    [print-rep ()
     ;* \ret{string}     
     ;* Builds a printable representation for the instance.
     (send self print-rep "")]
    [print-rep (str)
     ;* \ret{string}     
     ;* Builds a printable representation for the instance.
     (string-append "#i[" (symbol->string (class-name self))  " " str "]" )]

    [print ()
     ;* \ret{unspecified}
     ;* Displays a printable representation of the instance to the
     ;* current output port.
     (send self print (current-output-port))]
    [print (op)
     ;* \ret{unspecified}
     ;* Displays a printable representation of the instance to the
     ;* output port \var{op}.
     (fprintf op "~a" (send self print-rep))]
    
    [destroy ()
     ;* \ret{unspecified}
     ;* Destroys the object.

     ;; catching protected method calls here would entail banging in a
     ;; protected vector filled w/ error procs.  and we'd want to share
     ;; these (have one that was large enough for the biggest protected
     ;; table known)
     ;;
     ;; catching private method calls here is nigh-well impossible, thus
     ;; we only catch public calls
     (vector-set! self 0
       (lambda (msg instance . args)
         (unless
           (case msg
             ((destroy) (null? args))
             ((disown) (and (not (null? args)) (null? (cdr args))))
             (else #f))
           (error msg "instance has been destroyed"))))
     ;; experimental:  explicitly blitz all ivars (but not class prot table)
     ;; (nuking class prot table would make instance? unhappy)
     (let loop ((i (fx- (vector-length self) 1)))
       (unless (fx<= i 1)
         (vector-set! self i #f)
         (loop (fx- i 1))))]))

