;; Copyright (c) 1996 Oscar Waddell
;;
;; See the file "Notice" for information on usage and redistribution
;; of this file, and for a DISCLAIMER OF ALL WARRANTIES.

;; Todo
;;  - easy kills
;;     - pass in parent frame / toplevel or #f to create a toplevel so the dialog
;;       can be embedded in another window (change the destroy calls)
;;  - really need to implement a better geometry manager --- all these frames, yikes
;;  * find a way to avoid the following hacks used to center the window above its
;;    parent:
;;      - critical-section (needed so that update idletasks does something useful)
;;      - initial toplevel geometry of +2000+2000 (try to keep ugly stuff offscreen)
;;      - update idletasks

(swl:api-class (<option-button> parent default) (<proto-label> parent #f)
  ;; when button-1 is pressed, check to see if mouse was over button
  ;; if so, post menu  (can get event w/o mouse over button due to grab?)
  ;;
  ;; when posting menu, only bother if it actually contains entries
  ;;
  ;;
  ;% \swlapi{class}{<option-button>}{(create <option-button> parent)}
  ;* \ret{instance}
  ;* An option button allows the user to select an item from a list
  ;* of choices presented as a menu that appears over the button.
  ;* An option button displays the currently selected text string or image.
  ;*
  ;* Text is displayed in a uniform font (determined by \scheme{set-font!})
  ;* and is displayed as multiple lines if the text contains newlines
  ;* or its length exceeds the wrap length (see \scheme{set-wrap-length!}).
  ;* The text displayed by a button can be changed by the \scheme{set-title!}
  ;* method.
  ;*
  ;* Because the underlying widget library (Tcl/Tk) behaves differently
  ;* under Windows,
  ;* option buttons do not respond to the usual SWL event notification
  ;* methods.
  ;;
  (ivars (options '()) (menu #f) (tkvar (gensym)) (default default))
  (inherited handle)       ; handle is also used as the Tk var for the radiobutton menu entries
  (inheritable handle)
  (private)
  (protected)  ; no show-pressed or show-released  (no Tag bind for this guy)
  (public
    [init (parent default)
     ;% \ret{unspecified}
     (set! handle (swl:new-handle parent))
     ; note:  using handle for the varName attribute since radio-menu-items
     ; currently use the menu name as their -variable attribute...
     (swl:tcl-eval '|tk_optionMenu| handle tkvar default)
     (swl:tcl-eval 'set tkvar default)
     (send-base self init parent)
     ; restore Tk bindings:
     (swl:tcl-eval 'bindtags handle '|Menubutton|)]
    [set-options! (ls)
     ; hold menu w/ Scheme pointer
     ;*   ls is a list of strings or pairs whose car is a string and whose
     ;*   cdr is a procedure of no arguments
     (let ([m (create <menu>
                (map (lambda (opt)
                       (call-with-values
                         (lambda ()
                           (cond
                             [(pair? opt) (values (car opt) (cdr opt))]
                             [else (values opt #f)]))
                         (lambda (txt proc)
                           (unless (string? txt)
                             (error 'set-options! "~s is not a string" txt))
                           (unless (or (not proc) (procedure? proc))
                             (error 'set-options! "~s is not a procedure" proc))
                           (create <radio-menu-item> with
                             (title: txt)
                             (selected: (string=? txt default))
                             (action:
                               (lambda (item)
                                 (set-title! self txt)
                                 (when proc (proc))))))))
                     ls))])
       (send m set-parent! self)
       (set-tk-option '-menu m (lambda (x) #t) 'valid)
       (set! options ls)
       (set! menu m))]
    [get-options ()
     ;* The behavior when the list is mutated is undefined.
     options]
    [set-title! (txt)
     (send-base self set-title! txt)
     (swl:tcl-eval 'set tkvar txt)
     (void)]
    [set-active-background-color! (val)
     ;* \ret{unspecified}
     ;; not too sure about the merit of this method
     ;* Sets the background color used when the mouse cursor is over
     ;* the widget to \var{val},
     ;* which is either a symbol naming a color in
     ;* \mytt{/usr/lib/X11/rgb.txt} or an instance of \scheme{<rgb>}.
     (set-tk-option '-activebackground val swl:color? 'set-active-background-color!)]
    [get-active-background-color ()
     ;* \ret{see below}
     ;; not too sure about the merit of this method
     ;* Returns the background color used when the mouse cursor is
     ;* over the widget.
     ;* The value returned is either a symbol naming a color in
     ;* \mytt{/usr/lib/X11/rgb.txt} or an instance of \scheme{<rgb>}.
     (get-tk-option '-activebackground tk->color 'get-active-background-color)]
    [set-active-foreground-color! (val)
     ;* \ret{unspecified}
     ;; not too sure about the merit of this method
     ;* Sets the foreground color used when the mouse cursor is over
     ;* the widget to \var{val},
     ;* which is either a symbol naming a color in
     ;* \mytt{/usr/lib/X11/rgb.txt} or an instance of \scheme{<rgb>}.
     (set-tk-option '-activeforeground val swl:color? 'set-active-foreground-color!)]
    [get-active-foreground-color ()
     ;* \ret{see below}
     ;; not too sure about the merit of this method
     ;* Returns the foreground color used when the mouse cursor is
     ;* over the widget.
     ;* The value returned is either a symbol naming a color in
     ;* \mytt{/usr/lib/X11/rgb.txt} or an instance of \scheme{<rgb>}.
     (get-tk-option '-activeforeground tk->color 'get-active-foreground-color)]
    [set-disabled-foreground-color! (val)
     ;* \ret{unspecified}
     ;* Sets the foreground color used when the widget is disabled
     ;* to \var{val},
     ;* which is either a symbol naming a color in
     ;* \mytt{/usr/lib/X11/rgb.txt} or an instance of \scheme{<rgb>}.
     (set-tk-option '-disabled val swl:color? 'set-disabled-foreground-color!)]
    [get-disabled-foreground-color ()
     ;* \ret{see below}
     ;* Returns the foreground color used when the widget is disabled.
     ;* The value returned is either a symbol naming a color in
     ;* \mytt{/usr/lib/X11/rgb.txt} or an instance of \scheme{<rgb>}.
     (get-tk-option '-disabled tk->color 'get-disabled-foreground-color)]
    [set-enabled! (val)
     ;* \ret{unspecified}
     ;; with a little more care we might get this via inheritance
     ;; I didn't think it was worth adding lots of depth to the class tree
     ;; for just a few couple little methods
     ;* \var{val} is a boolean value indicating whether or not the widget is
     ;* enabled and responsive to user interaction.
     (set-tk-option '-state
       (and (boolean? val) (if val 'normal 'disabled))
       (lambda (x) x)
       'set-enabled!)]
    [get-enabled ()
     ;* \ret{boolean}
     ;* Returns a boolean value indicating whether or not the widget is
     ;* enabled and responsive to user interaction.
     (get-tk-option '-state (lambda (x) (not (string=? x "disabled"))) 'get-enabled)]))

; unfortunate aspect of using macros to rewrite classnames as refs
; to runtime class ids is that we get hosed by cyclic dependencies
(define isa-option-button? (lambda (x) (isa? x <option-button>)))

(swl:api-procedure swl:font-dialog
  ;* \formdef{swl:font-dialog}{procedure}{(swl:font-dialog parent families sizes styles current-font new-font cleanup)}
  ;* \ret{See below}
  ;* This procedure displays a font dialog
  ;* for the given \scheme{parent} (an instance of \scheme{<toplevel>} or \scheme{#f}, if none).
  ;* SWL attempts to place the dialog centered above the parent, if given, or
  ;* centered on the screen.
  ;* The \scheme{families}, \scheme{sizes}, and \scheme{styles} parameters are
  ;* lists of values appropriate for creating instances of the \scheme{<font>}
  ;* class.
  ;* The \scheme{current-font} procedure is invoked with no arguments, and should
  ;* return an instance of \scheme{<font>} used to supply default values for the
  ;* widgets displaying the various options.
  ;* When the ``OK'' button is pressed, the \scheme{new-font} procedure is invoked
  ;* with an instance of the selected font.
  ;* When the ``Cancel'' button is pressed, \scheme{new-font} is invoked with \scheme{#f}.
  ; get-current should return the current font
  ; return-selected is invoked with either #f or a font
  (lambda (parent families sizes styles get-current return-selected)
    (define fmt-family (lambda (x) (symbol->string x)))
    (define fmt-size
      (lambda (x)
        (format (if (negative? x) "~a pixels" "~a points") (abs x))))
    (define build-menu
      (lambda (parent options fmt default update)
        (let ([formatted (map fmt options)])
          (letrec ([mb
                    (create <option-button> parent (fmt default) with
                      (width/char: (apply max (map string-length formatted)))
                      (options:
                        (map (lambda (x) (cons (fmt x) (lambda () (update x))))
                             options)))])
            (pack mb (side: 'top))))))
    (unless (or (not parent) (isa? parent <toplevel>))
      (error 'swl:font-dialog "parent must be #f or an instance of <toplevel>"))
    (let* ([default-font (get-current)]
           [family (send default-font get-family)]
           [size (send default-font get-size)]
           [style (send default-font get-style)]
           [font default-font])
      (define build-check
        (lambda (parent value update)
          (let ([cb (create <checkbutton> parent with
                      (title: (symbol->string value))
                      (anchor: 'w)
                      (action: (lambda (b) (update value))))])
            (when (memq value style) (send cb invoke))
            (pack cb (side: 'top) (fill: 'x)))))
     (critical-section
      (let* ([top (create <toplevel> with (title: "Font dialog")
                    (transient: parent)
                    (geometry: "+2000+2000")
                    (resizable: #f #f))]
             [N (create <frame> top)]
             [S (create <frame> top)]
             [NW (create <frame> N)]
             [NE (create <frame> N with (relief: 'ridge) (border-width: 3))]
             [NWN (create <frame> NW)]
             [NWNW (create <frame> NWN with (relief: 'ridge) (border-width: 3))]
             [NWNE (create <frame> NWN with (relief: 'ridge) (border-width: 3))]
             [sample (create <entry> NW with (width/char: 10) (font: default-font))]
             [ok (create <button> S with (title: "Ok")
                   (action: (lambda (b) (destroy top) (return-selected font))))]
             [cancel (create <button> S with (title: "Cancel")
                       (action: (lambda (b) (destroy top) (return-selected #f))))]
             [sync (lambda ()
                     (set! font (create <font> family size style))
                     (set-font! sample font))])
        (define-syntax update
          (syntax-rules ()
            [(_ var val) (begin (set! var val) (sync))]))
        (define opposite
          (lambda (x)
            (case x
              [(bold) 'normal]
              [(normal) 'bold]
              [(italic) 'roman]
              [(roman) 'italic]
              [else #f])))
        (define-syntax toggle
          (syntax-rules ()
            [(_ val)
             (let ([flag #f])
               (lambda (b)
                 (set! flag (not flag))
                 (if flag
                     (set! style (cons val (remq val (remq (opposite val) style))))
                     (set! style
                       (remq val
                         (cond
                           [(opposite val) => (lambda (x) (cons x style))]
                           [else style]))))
                 (sync)))]))
      ; arrange for the font dialog to be destroyed / hidden / shown when the parent is
        (when parent (send parent add-follower top))
        (pack (create <label> NWNW with (title: "Family") (anchor: 'w)) (side: 'top))
        (build-menu NWNW families fmt-family family (lambda (x) (update family x)))
        (pack (create <label> NWNE with (title: "Size") (anchor: 'w)) (side: 'top))
        (build-menu NWNE sizes fmt-size size (lambda (x) (update size x)))
        (for-each
          (lambda (s)
            (when (memq s styles) (build-check NE s (toggle s))))
          '(bold italic overstrike underline))
        (pack N (side: 'top) (fill: 'x))
        (pack S (side: 'top))
        (pack ok (side: 'left))
        (pack cancel (side: 'left))
        (pack NW (side: 'left) (fill: 'y))
        (pack NE (side: 'right) (fill: 'y))
        (pack NWN (side: 'top) (fill: 'x))
        (send sample insert "Sample text")
        (pack sample (side: 'top) (fill: 'x))
        (pack NWNW (side: 'left) (fill: 'y))
        (pack NWNE (side: 'left) (fill: 'y))
        (with-values
          (if parent
              (values
                (get-root-x parent)
                (get-root-y parent)
                (send parent get-width)
                (send parent get-height))
              (values 0 0 (swl:screen-width) (swl:screen-height)))
          (lambda (x y width height)
            (swl:raw-tcl-eval "update idletasks")
            (send top set-geometry!
              (format "+~s+~s"
                (+ x (quotient (- width (send top get-width)) 2))
                (+ y (quotient (- height (send top get-height)) 2)))))))))))

