;; Copyright (c) 1996 Oscar Waddell
;;
;; See the file "Notice" for information on usage and redistribution
;; of this file, and for a DISCLAIMER OF ALL WARRANTIES.

;; call before doing any widget stuff

(define swl:with-library-file
 ; no idea why Tk insists on such ugly dialogs...  suppose I could cons up my own
 ; should provide graceful exodus, of course, but I have no time
  (lambda (name k)
    (define warn
      (lambda (title msg)
        (swl:tcl-eval 'tk_dialog '.swlerror title msg 'warning 0 "Drat")
        (format "library file ~s not found" name)))
    (let ([lib (getenv "SWL_ROOT")])
      (if (not lib)
          (warn "Configuration error"
            "Environment variable SWL_ROOT is not set.  Cannot locate library files.")
          (let ([path (string-append lib "/" name)])
            (if (file-exists? path)
                (k path)
                (warn "Installation error"
                  (format "Cannot locate SWL library file ~s in ~s" name lib))))))))

(define swl:error-log-port)
(define swl:show-balloon-help)
(define swl:hide-balloon-help)
(define initialize-graphics-lib
  (let ([destroyer
         (send (make <callback-proc>
                 (lambda (widget)
                   (let ([widget (swl:lookup-widget widget)])
                     (when widget (send widget destroy))))
                   #f #t) init)])
     (define swl:make-error-log-port
       (lambda ()
         (let ([top #f] [txt #f])
           (define make-new-top
             (lambda ()
               (set! top
                 (create <toplevel> with
                   (title: "SWL Error")
                   (min-size: 100 100)
                   (destroy-request-handler: (lambda (t) (set! top #f) #t))))
               (hide top)
               (let ([sf (create <scrollframe> top with (default-vscroll: #t) (sticky-hscroll: #t))]
                     [button
                      (create <button> top with
                        (title: "Dismiss")
                        (action: (lambda (b) (send top hide))))])
                 (pack sf (side: 'top) (fill: 'both))
                 (pack button (side: 'top))
                 (set! txt
                   (create <text> sf with
                     (height/char: 5)
                     (width/char: 70)
                     (wrap: 'none)
                     (background-color: 'white)))
                 (send txt set-enabled! #f)
                 (send txt show)
                 (show top))))
           (let ([op (open-output-string)])
             (define handler
               (lambda (xmsg . xargs)
                 (record-case (cons xmsg xargs)
                   [flush-output-port (p)
                    (critical-section
                      (let ([s (get-output-string op)])
                        (unless (string=? s "")
                          (unless top (make-new-top))
                          (send txt set-enabled! #t)
                          (send txt insert s)
                          (send txt set-enabled! #f)
                          (show top)
                          (raise top))))]
                   [write-char (c p) (write-char c op)]
                   [port-name (p) "swl:error-log-port"]
                   [else (let ((args (cdr xargs)))
                          (apply (port-handler op) xmsg op args))])))
             (make-output-port handler "")))))
    (lambda ()
      (swl:tcl-init 1 1)
      (swl:tcl-eval 'wm 'withdraw #\.)
     ; get tcl not to change focus on tab
      (swl:tcl-eval 'bind 'all '<|Key-Tab|> '())
      (swl:tcl-eval 'bind 'all '<|Shift-Key-Tab|> '())
      (swl:with-library-file "menu.tcl" (lambda (patch) (swl:tcl-eval 'source patch)))

      ;; hopefully this protects us somewhat from Tk's "send" security hole
      (swl:raw-tcl-eval "tk appname \"\"")
; doesn't exist under Windows NT (in Tk4.2)
;     (swl:raw-tcl-eval "rename send \"\"")

      ;; If we restore the binding for <Destroy>, modify the callback-proc
      ;; for destroyer (above) so that it returns swl:system-queue and a widget
      ;; handle to us.  Basically that means changing #f above to
      ;; "swl:system-queue %W".
      ;;;     (swl:tcl-eval 'bind '|SWL| '|<Destroy>| destroyer)

      ;; small hack here.
      ;; s_eval will return to us as a fallback whatever args we've indicated
      ;; here.  The first is the index of the destroy callback in the callback
      ;; table.  The second is a token used by swl:event-dispatch to dispatch
      ;; this callback to the appropriate queue. In this case, swl:system-queue
      ;; is a token that points us to a special queue where each fallback is
      ;; evaluated in a separate thread (so none can clog the system queue).
      ;; Finally, s_eval returns to us the handle of the window we need to
      ;; send the destroy message.
      (set! swl:make-destroy-notify
        (lambda (handle)
          (format "s_eval ~a swl:system-queue ~a"
            (send destroyer index-of) handle)))

      (let ((q (thread-make-msg-queue 'idle)))
        (thread-fork-group
         (lambda ()
           (thread-name "System Fallback")
           (parameterize ([current-output-port thread-conout]
                          [console-output-port thread-conout])
             (call/cc thread-become-server!)
             (let loop ()
               (let ([fb (thread-receive-msg q)])
                (thread-fork
                  (lambda () (swl:apply-callback-proc (car fb) (cdr fb))))
               (loop)))))
         (thread-default-quantum)
         -1 ;;; run with high priority
         )
         ; (maybe insert-widget isn't the best name in the world)
         (swl:insert-widget 'swl:system-queue #f (make-swl:fbq q 'swl:system-queue)))

      (let ([balloon #f] [thread #f])
         (define make-balloon
           (lambda ()
'            (let ([b (create <balloon>)])
               (send b hide)
               (swl:sync-display)
               (set! balloon b))))
        (set! swl:balloon-help-delay
          (make-parameter
            1000
            (lambda (x)
              (unless (and (fixnum? x) (not (fxnegative? x)))
                (error 'swl:balloon-help-delay "invalid delay ~s" x))
              x)))
        (set! swl:show-balloon-help
          (lambda (content x y width height)
(printf "balloon-help disabled for now~n")
'           (thread-critical-section
              (unless balloon (make-balloon))
              (when thread (thread-kill thread))
              (set! thread
                (thread-fork-group
                  (lambda ()
                    (thread-sleep (swl:balloon-help-delay))
                    (show balloon content x y width height)))))))
        (set! swl:hide-balloon-help
          (lambda ()
'           (thread-critical-section
              (when thread (thread-kill thread))
              (when balloon (send balloon hide))
              (set! thread #f)))))

      (swl:tcl-eval 'bind '|SWL| '|<Configure>| configure-cbproc)
      (swl:tcl-eval 'bind '|SWL| '|<Enter>| enter-cbproc)
      (swl:tcl-eval 'bind '|SWL| '|<Leave>| leave-cbproc)
      (swl:tcl-eval 'bind '|SWL| '|<KeyPress>| keypress-cbproc)
      (swl:tcl-eval 'bind '|SWL| '|<KeyRelease>| keyrelease-cbproc)
      (swl:tcl-eval 'bind '|SWL| '|<ButtonPress>| mousepress-cbproc)
      (swl:tcl-eval 'bind '|SWL| '|<ButtonRelease>| mouserelease-cbproc)
      (swl:tcl-eval 'bind '|SWL| '|<Double-ButtonPress>| (multi-mousepress-cbproc 2048))
      (swl:tcl-eval 'bind '|SWL| '|<Double-ButtonRelease>| (multi-mouserelease-cbproc 2048))
      (swl:tcl-eval 'bind '|SWL| '|<Triple-ButtonPress>| (multi-mousepress-cbproc 4096))
      (swl:tcl-eval 'bind '|SWL| '|<Triple-ButtonRelease>| (multi-mouserelease-cbproc 4096))
      (swl:tcl-eval 'bind '|SWL| '|<Motion>| mousemotion-cbproc)

      (set! swl:error-log-port
        (thread-make-parameter (swl:make-error-log-port)
          (lambda (x)
            (unless (output-port? x)
              (error 'swl:error-log-port "~s is not an output-port" x))
            x))))))

