;; Copyright (c) 1996 Oscar Waddell
;;
;; See the file "Notice" for information on usage and redistribution
;; of this file, and for a DISCLAIMER OF ALL WARRANTIES.

;;; expects to read boot file name from stdin

;;; borrowed from ../../nt/cat.ss
(define read-filename
  (lambda ()
    (let eat-whitespace ()
      (let ([c (peek-char)])
        (unless (eof-object? c)
          (when (char-whitespace? c) (read-char) (eat-whitespace)))))
    (let ([op (open-output-string)])
      (let lp ()
        (let ([x (read-char)])
          (case x
            [(#\space #\tab #\newline #!eof) (get-output-string op)]
            [(#\\) (write-char #\/ op) (lp)]
            [else (write-char x op) (lp)]))))))

(define boot-file (read-filename))

(define fremd-load-list '())
(define fremd-load
  (lambda (path)
    (define massage-path
     ; squashes out ., .., and empty components where possible
      (lambda (path return)
        (define split-path
          (lambda (path)
            (let ([end (string-length path)])
              (let f ((start 0) (next 0))
                (cond
                  [(fx= next end) (list (substring path start next))]
                  [(memv (string-ref path next) '(#\/ #\\))
                   (let ([next (fx+ next 1)])
                     (cons (substring path start next)
                           (f next next)))]
                  [else (f start (fx+ next 1))])))))
        (define reslash
          (lambda (path)
            (let ([ls (string->list path)])
              (list->string
                (if (memq (machine-type) '(i3nt ppcnt))
                    (substq #\\ #\/ ls)
                    (substq #\/ #\\ ls))))))
        (let loop ([ls (split-path path)] [new '()])
          (cond
            [(null? ls)
             (return (reslash (apply string-append (reverse (cdr new))))
                     (car new))]
            [(and (equal? (car ls) "/") (not (null? new)))
             (loop (cdr ls) new)]
            [(equal? (car ls) "./")
             (loop (cdr ls) new)]
            [(and (equal? (car ls) "../")
                  (not (null? new))
                  (equal? (car new) "/"))
             (loop (cdr ls) new)]
            [(and (equal? (car ls) "../")
                  (not (null? new))
                  (not (equal? (car new) "../")))
             (loop (cdr ls) (cdr new))]
            [else (loop (cdr ls) (cons (car ls) new))]))))
    (massage-path
      (string-append (current-directory) "/"
        (let ((n (string-length path)))
          (if (and (> n 2) (equal? (substring path (- n 3) n) ".ss"))
              (substring path 0 (- n 3))
              path)))
      (lambda (dir fn)
        (let ([path (string-append dir fn)])
          (unless (member path fremd-load-list)
            (parameterize ([current-directory dir])
              (parameterize ((warning-handler list) (reset-handler abort))
                (load (string-append fn ".ss"))
                (compile-file fn)))
            (set! fremd-load-list (cons path fremd-load-list))))))))

(load "../../config.scheme")

(for-each fremd-load '(
    "syntax"
    "porthack"
    "../threads/threads"
    "foreign"
    "io"
    "generics"
    "tkstream"
    "base1"
    "callback"
    "event"
    "option"
    "base2"
    "image"
    "proto"
    "label"
    "button"
    "frame"
    "entry"
    "scale"
    "scrollbar"
    "listbox"
    "canvas"
    "canvasitem"
    "text"
    "markup" ;;; put this after canvas and text
    "menu"   ;; after callback.ss due to swl:fbq
    "selection"
    "dialog"
    "teventloop" ;; loaded after callback.ss for sake of swl:fbqueue
    "../../apps/common/scrollframe"
    ;; init loaded after frame for sake of balloon help
    ;; init loaded after scrollframe for sake of swl:error-log-port
    ;; init loaded after callback.ss for sake of swl:fbqueue
    "init"
    "require"))

(define-syntax require
  (syntax-rules ()
    ((_ x . ignore) (eval-when (eval) (fremd-load x)))))

(fremd-load "../../apps/repl/repl.ss")
(collect 4)
(fremd-load "../../apps/edit/edit.ss")
(collect 4)
(fremd-load "../../apps/htmlview/htmlview.ss")
(collect 4)

(compile-file "build-script" "build-script.so")
(set! fremd-load-list (cons "build-script" fremd-load-list))

(set! fremd-load-list
  (map (lambda (x) (string-append x ".so"))
       (reverse fremd-load-list)))

;;; borrowed from ../../nt/cat.ss
(define cat-to
  (lambda (mode dest srcs)
    (define buf-size 4096)
    (printf "cat-to: ~s ~s ~s~n" mode dest srcs)
    (let ([op (open-output-file dest mode)]
          [buf (make-string buf-size)])
      (for-each
        (lambda (src)
          (call-with-input-file src
            (lambda (ip)
              (let lp ()
                (let ([n (block-read ip buf buf-size)])
                  (unless (eof-object? n) (block-write op buf n) (lp)))))))
        srcs)
      (close-output-port op))))

(cat-to 'replace boot-file fremd-load-list)

(if (memq (machine-type) '(ppcnt i3nt))
    (let ([op (open-output-file "cleanso.bat" 'replace)])
      (for-each
        (lambda (x) (fprintf op "del ~a~%" x))
        fremd-load-list)
      (close-output-port op))
    (let ([op (open-output-file "cleanso" 'replace)])
      (for-each
        (lambda (x) (fprintf op "/bin/rm ~a~%" x))
        fremd-load-list)
      (close-output-port op)))
